@@ -2,7 +2,7 @@ pub mod assertions;
 pub mod reporters;
 
 use anyhow::Result;
-use handlebars::Handlebars;
+use handlebars::{handlebars_helper, Handlebars};
 use httpstat::{httpstat, Config as HttpstatConfig};
 use httpstat::{Header, StatResult as HttpstatResult, Timing as HttpstatTiming};
 use serde::de::Deserializer;
@@ -43,10 +43,14 @@ pub struct RequestConfig {
 	///
 	/// Contents are rendered with [`mod@handlebars`].
 	pub data: Option<String>,
-	/// Key/Value pairs of headers to send with the request
+	/// List of headers to send with the request
 	///
 	/// Header values are rendered with [`mod@handlebars`].
-	pub headers: Option<HashMap<String, String>>,
+	pub headers: Option<Vec<Header>>,
+	/// Render raw headers from a template
+	///
+	/// Merged with [`headers`].
+	pub headers_template: Option<String>,
 	/// The url to request
 	///
 	/// The url string is rendered with [`mod@handlebars`].
@@ -63,6 +67,7 @@ impl Default for RequestConfig {
 			request_method: "GET".into(),
 			data: Default::default(),
 			headers: Default::default(),
+			headers_template: Default::default(),
 			url: Default::default(),
 			assertions: vec![AssertionConfig::Equal(RequestAssertionConfig {
 				skip: None,
@@ -102,6 +107,8 @@ pub struct Config {
 	pub connect_timeout: Option<u64>,
 	/// Verbose output
 	pub verbose: bool,
+	/// Fail on HTTP error response codes
+	pub fail_request: bool,
 	/// Maximum response size in bytes
 	pub max_response_size: Option<usize>,
 	/// Requests to run
@@ -342,6 +349,7 @@ where
 
 			Poll::Ready(Ok(results))
 		} else {
+			context.waker().wake_by_ref();
 			Poll::Pending
 		}
 	}
@@ -527,6 +535,9 @@ where
 	all_ready
 }
 
+handlebars_helper!(eqi: |x: String, y: String| x.to_lowercase() == y.to_lowercase());
+handlebars_helper!(nei: |x: String, y: String| x.to_lowercase() != y.to_lowercase());
+
 /// Run the request
 ///
 /// Handles set up of request configuration to call [`mod@httpstat()`] with and handle rendering of templated request
@@ -539,19 +550,20 @@ async fn run_request<C>(
 where
 	C: Serialize,
 {
-	let handlebars = Handlebars::new();
+	// TODO move Handlebars setup out of run_request
+	let mut handlebars = Handlebars::new();
+	handlebars.register_helper("eqi", Box::new(eqi));
+	handlebars.register_helper("nei", Box::new(nei));
 
 	// Render header values with the template context
-	let headers = match request_config.headers {
+	let mut headers = match request_config.headers {
 		Some(ref headers) => {
-			let mut rendered_headers: Vec<Header> = Vec::new();
+			let mut rendered_headers = Vec::new();
 
-			for (header, value) in headers {
-				let rendered_value =
-					handlebars.render_template(value, &*context.lock().unwrap())?;
+			for header in headers {
 				rendered_headers.push(Header {
-					name: header.into(),
-					value: rendered_value,
+					name: header.name.clone(),
+					value: handlebars.render_template(&header.value, &*context.lock().unwrap())?,
 				});
 			}
 
@@ -560,6 +572,29 @@ where
 		None => None,
 	};
 
+	if let Some(ref headers_template) = request_config.headers_template {
+		let rendered_headers_template =
+			handlebars.render_template(headers_template, &*context.lock().unwrap())?;
+
+		let mut rendered_headers = Vec::new();
+		for line in rendered_headers_template.lines() {
+			if let Some((name, value)) = line.split_once(':') {
+				rendered_headers.push(Header {
+					name: name.into(),
+					value: value.trim_start().into(),
+				});
+			}
+		}
+
+		let mut inner_headers = if let Some(headers) = headers {
+			headers
+		} else {
+			Vec::new()
+		};
+		inner_headers.extend_from_slice(&rendered_headers);
+		headers = Some(inner_headers);
+	}
+
 	let url =
 		Url::parse(&handlebars.render_template(&request_config.url, &*context.lock().unwrap())?)?;
 
@@ -582,7 +617,17 @@ where
 	};
 
 	match httpstat(&httpstat_config).await {
-		Ok(httpstat_result) => Ok((request_config, Arc::new(httpstat_result.into()))),
+		Ok(httpstat_result) => {
+			if config.fail_request && (400..600).contains(&httpstat_result.response_code) {
+				Err(Error::RequestError(
+					request_config,
+					format!("HTTP {}", httpstat_result.response_code),
+				)
+				.into())
+			} else {
+				Ok((request_config, Arc::new(httpstat_result.into())))
+			}
+		}
 		Err(error) => Err(Error::RequestError(request_config, error.to_string()).into()),
 	}
 }
